module Board (make_board, Board, terrain, tile, size, get_bases,
              reachable_neighbors, show_board, reachable_move,
              next_pos, reachable_directions, is_lethal
             ) where

import Types
import Data.Array

data Field = Field { terrain :: Terrain } deriving Eq

instance Show Field where
    showsPrec d (Field t) =
        showParen (d >= 10) $ showString $ show t

data Board = Board (Array (Coord,Coord) Field) deriving Eq

-- instance Read Board where
--     readsPrec _ = reads_Board
-- 
-- reads_Board :: ReadS Board
-- reads_Board s | widths_equal =
--     [(ba)]
-- reads_Board _ = []
--   where
--   ls = lines s
--   widths = map length ls
--   widths_equal = and $ map (== head widths) $ tail widths
--
-- readBoard :: [String] -> Board

instance Show Board where
    showsPrec _ bd =
         showString $ concat
           [[showT $ terrain $ tile bd (x,y) | x <- [tlx..brx]] ++ "\n" 
                                             | y <- [bry,bry-1..tly]]
      where
       ((tlx,tly),(brx,bry)) = size bd

make_board :: Coord -> Coord -> [String] -> Board
make_board w h l = Board . listArray ((1,1),(h,w)) $
                   map (Field . read . \a->[a]) $ concat l

tile :: Board -> (Coord, Coord) -> Field
tile (Board bd) (x,y) = bd ! (y,x)

size :: Board -> ((Coord, Coord), (Coord, Coord))
-- tlx = top left x, bry = bottom right y
size (Board bd) = let ((tlx,tly),(brx,bry)) = bounds bd
                  in ((tly,tlx),(bry,brx))

get_bases :: Board -> [Position]
get_bases (Board b) = map (switch . fst) $
                      filter ((Base ==) . terrain . snd) $ assocs $ b

switch :: (a, b) -> (b, a)
switch (a, b) = (b, a)

type BoardImage = Array Position Char

show_board :: Board -> Position -> [Position] -> String
show_board b p es = concat [[ image''!(y,x) | x <- [lo_x..hi_x]]
                            ++ ['\n'] | y <- [hi_y,hi_y-1..lo_y]]
    where
    ((lo_y,lo_x),(hi_y,hi_x)) = bounds image
    image   = paint_board b
    image'  = paint_enemies image $ map (validate.switch) es
    image'' = paint_postman image' $ switch p
    -- paint the board
    paint_board :: Board -> BoardImage
    paint_board (Board b') = listArray (bounds b')
                            $ map (showT . terrain) $ elems b'
    -- paint the enemies at the board image
    paint_enemies :: BoardImage -> [Position] -> BoardImage
    paint_enemies i e = i // (zip e $ repeat 'E')
    -- paint our hero at the board image
    paint_postman :: BoardImage -> Position -> BoardImage
    paint_postman i p' = i // [(p','P')]
    -- ensure the validity of a switched position, wrt image
    validate :: (Coord, Coord) -> (Coord, Coord)
    validate sp = if inRange (bounds image) sp then sp
                  else {- (1,hi_x) -} error $
                         "Wanted to paint at "++show sp++
                         " with bounds "++show (bounds image)

showT :: Terrain -> Char
showT Free = '.'
showT Base = '@'
showT Water = '~'
showT Wall = '#'

is_lethal :: Board -> Position -> Bool
is_lethal b p = (terrain (tile b p) == Water)

isfree :: Board -> Position -> Bool
isfree b p = isfree' $ terrain $ tile b p
    where
    isfree' :: Terrain -> Bool
    isfree' Free = True
    isfree' Base = True
    isfree' _    = False

isinbounds :: Board -> Position -> Bool
isinbounds (Board b) = inRange (bounds b) . switch

isok :: Board -> Position -> Bool
isok b p = isinbounds b p && isfree b p

nextp :: Position -> Direction -> Position
nextp (x,y) N = (x,y+1)
nextp (x,y) S = (x,y-1)
nextp (x,y) E = (x+1,y)
nextp (x,y) W = (x-1,y)

next_pos :: Position -> Direction -> Position
next_pos = nextp

reachable_move :: Board -> Position -> Direction -> Maybe Position
reachable_move b p d = if isok b p' then Just p' else Nothing
    where
    p' = nextp p d

reachable_neighbors :: Board -> Position -> [Position]
reachable_neighbors b p = filter (isok b)
                          [nextp p N, nextp p S, nextp p W, nextp p E]

reachable_directions :: Board -> Position -> [Direction]
reachable_directions b p = filter (isok_d) [N, S, W, E]
    where
    isok_d :: Direction -> Bool
    isok_d = isok b . nextp p
