> module Knowledge (Knowledge, init_knowledge, update_knowledge,
>                   carrying, max_weight, my_pos, my_packages, ps_here,
>                   ppls, ppls_lp, board, move_no, others, name,
>                   remember_bid_cmd, reboots,
>                   is_robot_at, backing_off,
>                   continuing, dh, no_dh, no_reboots
>                  ) where

This module represents our postman's knowledge about the world.
It exports an ADT Knowledge and associated accessors.

> import Dijkstra
> import Maybe
> import Types
> import Board (Board, get_bases)

> import Maybe (isNothing, fromJust)
> import List
>-- import Debug.Trace (trace)

=== Internal data structure ===


> data Knowledge
>   = K { name :: ID,
>         maxw :: Weight,             -- our maximum carrying capacity
>         money :: Money,
>         my_pos :: Position,           -- our current position
>         weight :: Weight,             -- how much we're carrying
>         my_ps :: [Package],           -- any packages we're carrying
>         ps_here :: [(Package, Distance)], -- packages at the current position
>         
>         board :: Board,
>         others :: [Robot],            -- Details of all other robots
>         ppls :: [(Position, Maybe Count)], -- potential paclocs
>         ppls_lp :: [(Position, Maybe Count)], -- potential paclocs (low priority)
>         continuing :: Bool, -- Should we continue to the same place?
>         move_no :: Int,               -- Move number
>         dh :: Maybe DijkstraHeading,
>         reboots :: Int, -- Number of kicks in a row
>         rebooted :: Bool, -- Have we kicked or been kicked this turn?
>         back_off_turns :: Int,     -- For how many turns to back off
>         last_cmd :: Maybe Command    -- We don't have one initially
>       }
>   deriving Show


=== Public interface ===


Are we in back-off mode?

> backing_off :: Knowledge -> Bool
> backing_off k = back_off_turns k > 0


Forget our previous heading.

> no_dh :: Knowledge -> Knowledge
> no_dh k = k { continuing = False, dh = Nothing }

> no_reboots :: Knowledge -> Knowledge
> no_reboots k = k { reboots = 0 }

Record in our knowledge that we've made a certain bid for a certain
command. Because our bids are always deducted from our money, regardless
of whether we can execute the command or not, we can subtract the bid price
right away.

> remember_bid_cmd :: Money -> Command -> Knowledge -> Knowledge
> remember_bid_cmd b c k = k { money = money k - (abs b) ,
>                              last_cmd = Just c
>                            }


Is there a robot at position p?

> is_robot_at :: Knowledge -> Position -> Bool
> is_robot_at k p = or $ map ((==p).bot_pos) (others k)


How much weight are we currently carrying?

> carrying :: Knowledge -> Weight
> carrying = weight


What is our maximum carrying capacity?

> max_weight :: Knowledge -> Weight
> max_weight = maxw


Yield the list of packages we're carrying.

> my_packages :: Knowledge -> [Package]
> my_packages = my_ps


Construct an initial knowledge from the data we're given in the
initialisation phase of the game.

> init_knowledge :: Board -> Configuration -> Knowledge
> init_knowledge b (i,w,m) = K { name = i,
>                                maxw = w,
>                                money = m,
>                                my_pos = (1,1),
>                                weight = 0,
>                                my_ps = [],
>                                ps_here = [],
>                                board = b,
>                                others = [],
>                                ppls = map (\x -> (x, Nothing)) $ get_bases b,
>                                ppls_lp = [],
>                                move_no = 0,
>                                continuing = False,
>                                last_cmd = Nothing,
>                                dh = Nothing,
>                                reboots = 0,
>                                rebooted = False,
>                                back_off_turns = 0
>                              }


Update our view of the world according to what happened during the last
turn. We send the original knowledge through a number of modifications:

  - inc_move just increments our internal move counter.
  - updates deals with anything that happened to the robots, including
    ourselves.
  - process_packages extracts knowledge from the new package list.

> update_knowledge :: Turn_Info -> Knowledge -> Knowledge
> update_knowledge (us, ps) k = k3
>   where k1 = k { rebooted = False }
>         k2 = if reboots k1 == 3 then k1 { back_off_turns = 3, reboots = 0 }
>                                 else k1
>         k3 = (counters . discover_reboots us . process_packages ps . updates us) k2

> counters :: Knowledge -> Knowledge
> counters k = if bt == 0 && rebooted k then k' { reboots = reboots k' + 1 }
>                                       else k'
>   where k' = k { move_no = move_no k + 1 ,
>                  back_off_turns = if bt>0 then bt-1 else 0
>                }
>         bt = back_off_turns k

> discover_reboots :: [Update] -> Knowledge -> Knowledge
> discover_reboots us k
>   = if what_we_wanted (map snd $ filter interesting us) k
>     then k else k { rebooted = True }
>   where interesting :: Update -> Bool
>         interesting (_, HPos _) = False
>         interesting (_, Alive) = False
>         interesting (i, _) = i == my_name
>         my_name = name k

> what_we_wanted :: [Happened] -> Knowledge -> Bool
> what_we_wanted hs k = case last_cmd k of
>                         Just Dunno -> True
>                         Just lc -> set_equal hs $ case lc of
>                               Move N -> [HN]
>                               Move S -> [HS]
>                               Move E -> [HE]
>                               Move W -> [HW]
>                               Pick is -> map HP is
>                               Drop is -> map HD is
>                               _ -> hs -- Can't happen
>                         Nothing -> True

> set_equal :: Eq a => [a] -> [a] -> Bool
> set_equal xs ys = null (xs\\ys) && null (ys\\xs)

This is what we get from the package info:

  - The packages at our current location (ps_here).
  - If we encounter any packages, we remember this spot as a
    possible package location. Otherwise we strike it from both
    ppl lists.

> process_packages :: [Package] -> Knowledge -> Knowledge
> process_packages ps k = let k' = k { ps_here = useful_ps_dists,
>                             ppls = if null ps
>                                    then ppls'
>                                    else (mp, Just (length useful_ps)):ppls',
>                             ppls_lp = ppls_lp' }
>                         in if null ps && may_have_packages
>                            then no_dh k'
>                            else k'
>   where ppls' = filter ((mp /=) . fst) (ppls k)
>         ppls_lp' = filter ((mp /=) . fst) (ppls_lp k)
>         may_have_packages = isJust $ find ((mp ==) . fst) (ppls k ++ ppls_lp k)
>         mp = my_pos k
>         carryable_ps = filter ((max_weight k >=) . package_weight) ps
>         b = board k
>         useful_ps_dists = stripNothing $ zip carryable_ps (distanceToPoints b [] mp (map package_pos carryable_ps))
>         useful_ps = map fst useful_ps_dists

> stripNothing :: [(a, Maybe b)] -> [(a, b)]
> stripNothing = map (\(a,b) -> (a, fromJust b)) . filter (isJust . snd)

Robot updates give us the following:

  - We can deduce the set of live robots. update_alives does that; it
    throws away any dead robots and adds entries for any new ones.
  - update_stuffs deals with updates happening to other robots.
  - self_update processes updates to our own robot.

> updates :: [Update] -> Knowledge -> Knowledge
> updates us k = update_stuffs stuff $
>                update_alives alives k
>   where alives = map fst $ filter (\(i, h) -> i /= me && h == Alive) us
>         stuff = filter ((Alive /=) . snd) us
>         me = name k


New robots get a dummy position of (0,0). This would break our program
should it not be updated to a valid position, but we trust the server
to have sent an XY message for such bots.

> update_alives :: [ID] -> Knowledge -> Knowledge
> update_alives is k = k { others = others'' }
>   where is' = sort is
>         others' = filter ((`elem` is) . bot_id) (sortBy compare_robots $ others k)
>         others''
>           = merge compare_robots others'
>             (map (\i -> Robot {bot_id = i, bot_pos = (0, 0), bot_count = 0}) is')
>             --(map (\i -> (i, (0, 0), 0)) is') -- Random position!

> compare_robots :: Robot -> Robot -> Ordering
> compare_robots b1 b2 = compare (bot_id b1) (bot_id b2)

> merge :: (a -> a -> Ordering) -> [a] -> [a] -> [a]
> merge _ [] ys = ys
> merge _ xs [] = xs
> merge f (x:xs) (y:ys) = case f x y of
>                             LT -> x:merge f xs (y:ys)
>                             EQ -> x:merge f xs ys
>                             GT -> y:merge f (x:xs) ys


update_stuffs just repeatedly updates our knowledge with a single update.

> update_stuffs :: [Update] -> Knowledge -> Knowledge
> update_stuffs [] k = k
> update_stuffs us k = foldl update_stuff k us

Robot movements are recorded in our knowledge via the function
update_stuff.

> update_stuff :: Knowledge -> Update -> Knowledge
> update_stuff k (i, h)
>   = if i == name k
>     then upd k h
>     else case h of
>            HN -> apply_position_update i (\(x, y) -> (x, y + 1)) k
>            HS -> apply_position_update i (\(x, y) -> (x, y - 1)) k
>            HE -> apply_position_update i (\(x, y) -> (x + 1, y)) k
>            HW -> apply_position_update i (\(x, y) -> (x - 1, y)) k
>            HPos (x, y) -> apply_position_update i (\_ -> (x, y)) k
>            HP _ -> picked i k
>            HD _ -> dropped i k
>            _ -> k -- error "update_stuff"

> dropped :: ID -> Knowledge -> Knowledge
> dropped i k = case find ((i ==) . bot_id) (others k) of
>                   Just drop_bot -> let 
>                                    pos = bot_pos drop_bot
>                                    in
>                                    if any ((pos ==) . fst) ppls'
>                                    then k' { ppls = inc_pos pos ppls' }
>                                    else if any ((pos ==) . fst) ppls_lp'
>                                         then k' { ppls_lp = inc_pos pos ppls_lp' }
>                                         else k' { ppls_lp = (pos, Just 1):ppls_lp' }
>                   Nothing -> k
>   where ppls' = ppls k
>         ppls_lp' = ppls_lp k
>         k' = no_dh $ dec_bot i k

> picked :: ID -> Knowledge -> Knowledge
> picked i k
>             = case find ((i ==) . bot_id) (others k) of
>                   Just pick_bot -> let
>                                    pos = bot_pos pick_bot
>                                    in
>                                    no_dh $ k' { ppls_lp = dec_pos pos ppls_lp',
>                                                 ppls = dec_pos pos ppls',
>                                               }
>                   Nothing -> k
>   where ppls' = ppls k
>         ppls_lp' = ppls_lp k
>         k' = inc_bot i k

> inc_pos :: Position -> [(Position, Maybe Count)] -> [(Position, Maybe Count)]
> inc_pos p pcs = case maybe_this_maybe_count of
>                     Just (Just c) -> (p, Just (c + 1)):other_poss
>                     Just Nothing -> pcs
>                     Nothing -> (p, Just 1):pcs
>   where maybe_this_maybe_count = lookup p pcs
>         other_poss = filter ((p /=) . fst) pcs

> dec_pos :: Position -> [(Position, Maybe Count)] -> [(Position, Maybe Count)]
> dec_pos p pcs = case maybe_this_maybe_count of
>                     Just (Just 1) -> other_poss
>                     Just (Just c) -> (p, Just (c - 1)):other_poss
>                     Just Nothing -> pcs
>                     Nothing -> pcs
>   where maybe_this_maybe_count = lookup p pcs
>         other_poss = filter ((p /=) . fst) pcs

> dec_bot :: ID -> Knowledge -> Knowledge
> dec_bot = alter_robot (+ (-1))

> inc_bot :: ID -> Knowledge -> Knowledge
> inc_bot = alter_robot (+ 1)

> alter_robot :: (Count -> Count) -> ID -> Knowledge -> Knowledge
> alter_robot f i k = case maybe_this_robot of
>                         Just _ -> k { others = this_robot:other_robots }
>                         Nothing -> k -- Can't happen, but...
>   where others' = others k
>         other_robots = filter ((i /=) . bot_id) others'
>         maybe_this_robot = find ((i ==) . bot_id) others'
>         this_robot = case fromJust maybe_this_robot of
>                          --(j, p, c) -> (j, p, f c)
>                          b1 -> b1 { bot_count = f (bot_count b1) }

> apply_position_update :: ID -> (Position -> Position)
>                       -> Knowledge -> Knowledge
> apply_position_update i f k = k { others = u (others k) }
>   where
>   u :: [Robot] -> [Robot]
>   u [] = []
>   u (b:bs)
>     | bot_id b==i  = b {bot_pos = f (bot_pos b)} : bs
>     | otherwise    = b : u bs


self_update does the obvious for movements and position update,
but also deals with package pick and drop by adding/remoning the package
question to/from our package list. It also updates the weight we're
carrying accordingly.

> upd :: Knowledge -> Happened -> Knowledge
> upd k h
>     = case h of
>       HN -> addy 1 k
>       HS -> addy (-1) k
>       HW -> addx (-1) k
>       HE -> addx 1 k
>       HPos p -> k {my_pos = p}
>       HP i -> case find_package i $ map fst $ ps_here k of
>                       Just p -> k { my_ps = (p:my_ps k),
>                                     weight = weight k + package_weight p }
>                       Nothing -> k -- error $ "upd: find_package " ++
>                                    -- show i++" from "++show (ps_here k)
>       HD i -> let ps = my_ps k
>                   (p,ps') = drop_package i ps
>                   in k { my_ps = ps',
>                          weight = weight k - package_weight p }
>       Alive -> k -- error "upd: We don't want Alive here!"
> addy, addx :: Int -> Knowledge -> Knowledge
> addy n k = let (x,y) = my_pos k in k {my_pos = (x,y+n)}
> addx n k = let (x,y) = my_pos k in k {my_pos = (x+n,y)}

> find_package :: ID -> [Package] -> Maybe Package
> find_package _ [] = Nothing
> find_package i (p:ps)
>   | package_id p == i = Just p
>   | otherwise         = find_package i ps

> drop_package :: ID -> [Package] -> (Package, [Package])
> drop_package _ [] = ((0, (0, 0), 0), []) -- error "drop_package"
> drop_package i (p:ps)
>   | package_id p == i = (p,ps)
>   | otherwise         = let (p',ps') = drop_package i ps in (p', (p:ps'))

