module Tests where

import AI
import Board
import Dijkstra
import IO
import Knowledge
import Maybe
import Monad
import QuickCheck
import Types
-- import Debug.Trace

instance Arbitrary Terrain where
    arbitrary = frequency
                [(10,return Free),
                 (5, return Water),
                 (7, return Wall),
                 (1, return Base)]
    coarbitrary = error "not implemented"

-- obviously, this is not a complete definition for how to make random values of type Char
instance Arbitrary Char where
    arbitrary = oneof (map return (['.','~','#','@']))
    coarbitrary = variant . fromEnum

prop_ReadTerrain :: Char -> Property
prop_ReadTerrain c = elem c ['.','~','#','@'] ==> [c] == showT (read [c])

prop_ShowTerrain :: Terrain -> Bool
prop_ShowTerrain t = t == read (showT t)

-- imho, this should become toplevel & exported in Board
showT :: Terrain -> [Char]
showT Free = "."
showT Base = "@"
showT Water = "~"
showT Wall = "#"

maxboardsize :: Int
maxboardsize = 50

vector' :: Int -> Gen a -> Gen [a]
vector' n gen = sequence (replicate n gen)

-- I suspect I won't understand this tomorrow
instance Arbitrary Board where
    arbitrary = do width <- choose (1,maxboardsize)
                   height <- choose (1,maxboardsize)
                   let tile_freq = frequency [(10,return '.'),
                                              (5,return '~'),
                                              (7,return '#'),
                                              (1,return '@')]
                   row_list <- vector' height (vector' width tile_freq)
                   return $ make_board width height row_list
    coarbitrary = error "not implemented"

getB :: Gen Board -> IO ()
getB genboard = do 
                board <- genboard
                h <- openFile "random.map"
                hPutStrLn h (show board)
                hClose h
        

instance Arbitrary SearchType where
    arbitrary = frequency [(1,return FixedStart),(1,return FixedEnd)]
    coarbitrary = error "not implemented"

newtype PathProblem = PP (Board,[Position],SearchType,Position,Position)
  deriving Show

instance Arbitrary PathProblem where
    arbitrary = do width <- choose (1,maxboardsize)
                   height <- choose (1,maxboardsize)
                   let tile_freq = frequency [(10,return '.'),
                                              (5,return '~'),
                                              (7,return '#'),
                                              (1,return '@')]
                   row_list <- vector' height (vector' width tile_freq)
                   let board = make_board width height row_list
                   let pos = liftM2 (,) (choose (1,width)) (choose (1,height))
                   from <- pos
                   to <- pos
                   st <- arbitrary
                   fpslength <- choose (0,5 :: Int) 
                   fps <- vector' fpslength pos
                   return $ PP (board,fps,st,from,to)
    coarbitrary = error "not implemented"

-- this is just elements, I think
randItem :: [a] -> Gen a                       
randItem xs = do i <- choose (0, length xs-1)
                 return (xs!!i)

foo :: Gen [Char]
foo = vector 10 

prop_RSBoard   :: Board -> Bool
prop_RSBoard b = b == (make_board width height $ reverse $ lines $ show b)
                 where
                 (_,(brx,bry)) = size b 
                 height = bry
                 width = brx


tsp :: PathProblem -> Maybe (Distance,Path Position)
tsp (PP (b,fps,st,from,to)) = snd $ testshortestpath b fps st from to

prop_ValidCells :: PathProblem -> Property
prop_ValidCells pp@(PP (b,_fps,_st,from,to))
 = validcell b from && validcell b to ==> 
     maybe True (all (validcell b)) (fmap snd $ tsp pp)

prop_ValidPaths :: PathProblem -> Bool
prop_ValidPaths pp@(PP (b,_fps,_st,from,to))
 = validpath b from to $ fmap snd $ tsp pp

prop_EndsCorrectly :: PathProblem -> Bool
prop_EndsCorrectly pp@(PP (_b,_fps,_st,from,to))
 = maybe True (\ps -> last (from:ps) == to) (fmap snd $ tsp pp)

prop_SaneDistance :: PathProblem -> Bool
prop_SaneDistance pp@(PP (_b,_fps,_st,(fromx,fromy),(tox,toy)))
 = maybe True 
         (\d -> d >= 99 * (abs (tox-fromx) + abs (toy-fromy)))
         (fmap fst $ tsp pp)

prop_CorrectDistance :: PathProblem -> Bool
prop_CorrectDistance pp@(PP (_b,fps,_st,from,_))
 = maybe True
         (\(d,ps) -> d == distance fps (from:ps))
         (tsp pp)

distance :: [Position] -> [Position] -> Distance
distance _ [] = error "empty distance in Tests.hs"
distance _fps [_pos] = 0
distance fps (pos1:pos2:poss) 
  = (if elem pos1 fps || elem pos2 fps then 99 else 100)
  + distance fps (pos2:poss)


validcell :: Board -> Position -> Bool
validcell board (x,y) = tlx <= x && x <= brx && tly <= y && y <= bry
                          && (ter == Free || ter == Base)
                          where ((tlx,tly),(brx,bry)) = size board
                                ter  = terrain $ tile board (x,y)

validpath :: Board -> Position -> Position -> Maybe (Path Position) -> Bool
validpath _ _ _ Nothing = True
validpath _board from _to (Just ps) = (all validneighbour $ zip (from:ps) ps) 
    where validneighbour (x,y) = isJust (calc_direction x y)

-- distance tests

newtype DistanceProblem = DP (Board, Position, [Position])
    deriving Show

instance Arbitrary DistanceProblem where
    arbitrary = do width <- choose (1,maxboardsize)
                   height <- choose (1,maxboardsize)
                   let tile_freq = frequency [(10,return '.'),
                                              (5,return '~'),
                                              (7,return '#'),
                                              (1,return '@')]
                   row_list <- vector' height (vector' width tile_freq)
                   let board = make_board width height row_list
                   let pos = liftM2 (,) (choose (1,width)) (choose (1,height))
                   from <- pos
                   targslength <- choose (0,5 :: Int) 
                   targs <- vector' targslength pos
                   return $ DP (board,from,targs)
    coarbitrary = error "not implemented"

-- instance Arbitrary Position where
--     arbitrary = (choose (1,1000) , choose (1,1000) )
--quickcheck defines an instance for double, triples, etc

instance Arbitrary Robot where
    arbitrary = liftM3 Robot (choose (1, oneBil)) (arbitrary :: Gen Position) (choose (1,maxBound :: Int))

oneBil = 1000000000 :: Int

-- instance Arbitrary Knowledge where
--     arbitrary = do
--                 id <- choose (1, oneBil)
--                 maxw <- choose (1, oneBil))
--                 money <- choose (1, oneBil))
--                 my_pos <- (choose (1,1000),choose (1,1000))
--                 weight <- choose (1, oneBil))
--                 my_ps <- (choose (1, oneBil)),(choose (1,1000),choose (1,1000)),choose (1, oneBil)))
--                 board <- arbitrary :: Gen Board
                

                   
{-
distanceToPointsSimple :: Board -> Position -> [Position] -> [Maybe Distance]

distanceToPoints :: Board -> [Position] 
                        -> Position -> [Position] -> [Maybe Distance]
-}

testincdijkstra 
 = let 
       farpoint = (70,70)
       farpoint2 = (69,70)
       emptyboard = make_board 300 300 (repeat ".")
       Just (dir1,dh1) = best_step emptyboard [] Nothing (30,30) [farpoint]
       Just (dir2,dh2) = next_step emptyboard [] (30,31) dh1
       Just (dir3,dh3) = next_step emptyboard [] (29,31) dh2
       Just (dir4,dh4) = next_step emptyboard [] (29,32) dh3
       Just (dir5,dh5) = next_step emptyboard [] (28,32) dh4
       Just (dir6,dh6) = next_step emptyboard [] (28,33) dh5
       Just (dir7,dh7) = best_step emptyboard [] (Just dh6) (28,34) [farpoint]
       Just (dir8,dh8) = next_step emptyboard [] (28,35) dh7
       Just (dir9,dh9) = next_step emptyboard [] (27,34) dh8
       Just (dir10,dh10) = best_step emptyboard [] (Just dh9) (27,35) 
                                                         [farpoint,farpoint2]
       Just (dir11,dh11) = next_step emptyboard [] (27,36) dh10
       Just (dir12,dh12) = next_step emptyboard [] (26,36) dh11
   in do hSetBuffering stdout NoBuffering
         putStr "slow  : " >> print dir1
         putStr "fast  : " >> print dir2
         putStr "slow  : " >> print dir3
         putStr "fast  : " >> print dir4
         putStr "medium: " >> print dir5
         putStr "fast  : " >> print dir6
         putStr "slow  : " >> print dir7
         putStr "fast  : " >> print dir8
         putStr "medium: " >> print dir9
         putStr "slow  : " >> print dir10
         putStr "fast  : " >> print dir11
         putStr "slow  : " >> print dir12
